# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for DENY policies
#
# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-locals

import os
import time
import unittest

import pytest

from okera import _thrift_api
from okera.tests import pycerebro_test_common as common
from okera._thrift_api import TConfigType, TPrivilegeType

TEST_DB = 'deny_test_db'
TEST_TBL = 'tbl'
TEST_ROLE = 'deny_test_role'
TEST_USER = 'deny_test_user'

class DenyTest(common.TestBase):

    @classmethod
    def setUpClass(cls):
        """ Initializes one time state that is shared across test cases. This is used
            to speed up the tests. State that can be shared across (but still stable)
            should be here instead of __cleanup()."""
        super(DenyTest, cls).setUpClass()
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.delete_attribute('deny_test', 'v1')
            conn.create_attribute('deny_test', 'v1')
            conn.delete_attribute('deny_test', 's1')
            conn.create_attribute('deny_test', 's1')

    @staticmethod
    def __cleanup(conn, dbs=None):
        """ Cleanups all the test state used in this test to "reset" the catalog.
            dbs can be specified to do the initialize over multiple databases.
            This can be used for tests that use multiple dbs (but makes the test
            take longer). By default, only load TEST_DB.
        """
        conn.execute_ddl("DROP ROLE IF EXISTS %s" % TEST_ROLE)
        conn.execute_ddl("CREATE ROLE %s" % TEST_ROLE)
        conn.execute_ddl("GRANT ROLE %s to GROUP %s" % (TEST_ROLE, TEST_USER))

        conn.execute_ddl("DROP DATABASE IF EXISTS %s CASCADE" % TEST_DB)
        if not dbs:
            dbs = [TEST_DB]

        for db in dbs:
            conn.execute_ddl("CREATE DATABASE %s" % db)
            conn.execute_ddl("CREATE TABLE %s.%s(col1 int, col2 int, col3 int)" % \
                             (db, TEST_TBL))
            common.TestBase._create_all_types(conn, db)
            conn.assign_attribute('deny_test', 'v1', db, 'alltypes', 'int_col')
            conn.assign_attribute('deny_test', 'v1', db, 'alltypes', 'string_col')
            conn.assign_attribute('deny_test', 's1', db, 'alltypes', 'string_col')

    def test_deny_ddl(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__cleanup(conn)
            conn.assign_attribute('deny_test', 'v1', TEST_DB, TEST_TBL, 'col1')

            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.execute_ddl(
                    "DENY SELECT ON DATABASE %s FROM ROLE %s" % (TEST_DB, TEST_ROLE))
            self.assertTrue('must specify attribute expression' in str(ex_ctx.exception),
                            msg=str(ex_ctx.exception))

            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.execute_ddl(
                    ("DENY SELECT ON DATABASE %s HAVING ATTRIBUTE %s WHERE %s " +
                     "FROM ROLE %s") % \
                    (TEST_DB, "deny_test.v1", "true", TEST_ROLE))
            self.assertTrue('do not support filters' in str(ex_ctx.exception),
                            msg=str(ex_ctx.exception))

            # This should fail, this user has no privileges
            ctx.enable_token_auth(token_str=TEST_USER)
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                data = conn.scan_as_json('%s.%s' % (TEST_DB, TEST_TBL))
            self.assertTrue('does not have privilege' in str(ex_ctx.exception),
                            msg=str(ex_ctx.exception))

            # Create a DENY policy
            ctx.disable_auth()
            conn.execute_ddl(
                "DENY SELECT ON DATABASE %s HAVING ATTRIBUTE %s FROM ROLE %s" % \
                (TEST_DB, "deny_test.v1", TEST_ROLE))
            result = conn.execute_ddl("SHOW GRANT ROLE %s" % TEST_ROLE)
            self.assertEqual(len(result), 1)
            self.assertEqual(len(result[0]), 16)
            self.assertTrue(result[0][13])

            # This should fail, this user has no grants, only denys
            ctx.enable_token_auth(token_str=TEST_USER)
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                data = conn.scan_as_json('%s.%s' % (TEST_DB, TEST_TBL))
            self.assertTrue('does not have privilege' in str(ex_ctx.exception),
                            msg=str(ex_ctx.exception))

            # Grant SELECT on the DB
            ctx.disable_auth()
            conn.execute_ddl(
                "GRANT SELECT ON DATABASE %s TO ROLE %s" % \
                (TEST_DB, TEST_ROLE))

            ctx.enable_token_auth(token_str=TEST_USER)
            data = conn.execute_ddl('describe %s.%s' % (TEST_DB, TEST_TBL))
            print(data)

    def test_restrict_ddl(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__cleanup(conn)

            # Negative cases: missing transform and filter
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.execute_ddl(
                    "RESTRICT SELECT ON DATABASE %s FROM ROLE %s" % (TEST_DB, TEST_ROLE))
            self.assertTrue('RESTRICT privileges must specify' in str(ex_ctx.exception),
                            msg=str(ex_ctx.exception))

            # Having restrict
            conn.execute_ddl(
                "RESTRICT SELECT ON DATABASE %s HAVING ATTRIBUTE %s FROM ROLE %s" % \
                (TEST_DB, "deny_test.v1", TEST_ROLE))
            result = conn.execute_ddl("SHOW GRANT ROLE %s" % TEST_ROLE)
            print(result)
            self.assertEqual(len(result), 1)
            self.assertEqual(len(result[0]), 16)
            self.assertTrue(result[0][13])

            params = _thrift_api.TGetRoleParams()
            params.role_name = TEST_ROLE
            params.include_privilege_props = True
            params.include_role_history = True
            result = conn._underlying_client().GetRole(params)
            role = result.role
            self.assertEqual(len(role.privileges), 1)
            priv = role.privileges[0]
            self.assertEqual(priv.privilege_type, TPrivilegeType.RESTRICT)
            history = result.history
            self.assertTrue(len(history) > 0)

            # History should be ordered (descending chronological order)
            ts = float("inf")
            for e in history:
              self.assertTrue(ts >= e.timestamp_ms)
              ts = e.timestamp_ms

            # Transform restrict
            conn.execute_ddl(
                "RESTRICT SELECT ON DATABASE %s %s FROM ROLE %s" % \
                (TEST_DB, "TRANSFORM deny_test.v1 WITH mask()", TEST_ROLE))
            result = conn.execute_ddl("SHOW GRANT ROLE %s" % TEST_ROLE)
            print(result)
            self.assertEqual(len(result), 2)
            self.assertEqual(len(result[0]), 16)
            self.assertTrue(result[0][13])
            self.assertEqual(len(result[1]), 16)
            self.assertTrue(result[1][13])

            # Filter restrict
            conn.execute_ddl(
                "RESTRICT SELECT ON DATABASE %s WHERE %s FROM ROLE %s" % \
                (TEST_DB, "true", TEST_ROLE))
            result = conn.execute_ddl("SHOW GRANT ROLE %s" % TEST_ROLE)
            self.assertEqual(len(result), 3)
            self.assertEqual(len(result[0]), 16)
            self.assertTrue(result[0][13])
            self.assertEqual(len(result[1]), 16)
            self.assertTrue(result[1][13])
            self.assertEqual(len(result[2]), 16)
            self.assertTrue(result[2][13])

            # Transform and filter
            conn.execute_ddl(
                "RESTRICT SELECT ON DATABASE %s %s WHERE %s FROM ROLE %s" % \
                (TEST_DB, "TRANSFORM deny_test.v1 WITH mask()", "true", TEST_ROLE))
            result = conn.execute_ddl("SHOW GRANT ROLE %s" % TEST_ROLE)
            self.assertEqual(len(result), 4)
            self.assertEqual(len(result[0]), 16)
            self.assertTrue(result[0][13])
            self.assertEqual(len(result[1]), 16)
            self.assertTrue(result[1][13])
            self.assertEqual(len(result[2]), 16)
            self.assertTrue(result[2][13])
            self.assertEqual(len(result[3]), 16)
            self.assertTrue(result[3][13])

            result = conn.execute_ddl_table_output("SHOW GRANT ROLE %s" % TEST_ROLE)
            print(result)


            # Having - drop restrict
            conn.execute_ddl(
                "DROP RESTRICT SELECT ON DATABASE %s HAVING ATTRIBUTE %s FROM ROLE %s" % \
                (TEST_DB, "deny_test.v1", TEST_ROLE))

            # Transform - drop restrict
            conn.execute_ddl(
                "DROP RESTRICT SELECT ON DATABASE %s %s FROM ROLE %s" % \
                (TEST_DB, "TRANSFORM deny_test.v1 WITH mask()", TEST_ROLE))

            # Filter - drop restrict
            conn.execute_ddl(
                "DROP RESTRICT SELECT ON DATABASE %s WHERE %s FROM ROLE %s" % \
                (TEST_DB, "true", TEST_ROLE))

            # Transform and filter - drop restrict
            conn.execute_ddl(
                "DROP RESTRICT SELECT ON DATABASE %s %s WHERE %s FROM ROLE %s" % \
                (TEST_DB, "TRANSFORM deny_test.v1 WITH mask()", "true", TEST_ROLE))

            print("*************** GRANTS after DROP RESTRICT statements: ***********")
            result = conn.execute_ddl("SHOW GRANT ROLE %s" % TEST_ROLE)
            print(result)
            self.assertTrue(not result)

    def test_drop_restrict_ddl(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__cleanup(conn)

            #### Case 1: ####
            # Try to drop a non-existing restrict privilege, it should throw exception
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.execute_ddl(
                    "DROP RESTRICT SELECT ON DATABASE %s HAVING ATTRIBUTE %s FROM ROLE %s" % \
                    (TEST_DB, "deny_test.v1", TEST_ROLE))
            self.assertTrue("CatalogException: Grant WITH EXPRESSION 'deny_test.v1' does not exist."
                            in str(ex_ctx.exception), msg=str(ex_ctx.exception))

            # Now, create the restrict privilege
            conn.execute_ddl(
                "RESTRICT SELECT ON DATABASE %s HAVING ATTRIBUTE %s FROM ROLE %s" % \
                (TEST_DB, "deny_test.v1", TEST_ROLE))
            result = conn.execute_ddl("SHOW GRANT ROLE %s" % TEST_ROLE)
            self.assertEqual(len(result), 1)
            self.assertEqual(len(result[0]), 16)
            self.assertTrue(result[0][13])

            # Try to drop newly created restrict privilege, it should be successful
            conn.execute_ddl(
                "DROP RESTRICT SELECT ON DATABASE %s HAVING ATTRIBUTE %s FROM ROLE %s" % \
                (TEST_DB, "deny_test.v1", TEST_ROLE))
            result = conn.execute_ddl("SHOW GRANT ROLE %s" % TEST_ROLE)
            self.assertTrue(not result)



            #### Case 2: ####
            # Create a new restrict privilege with : Transform and filter
            conn.execute_ddl(
                "RESTRICT SELECT ON DATABASE %s %s WHERE %s FROM ROLE %s" % \
                (TEST_DB, "TRANSFORM deny_test.v1 WITH mask()", "true", TEST_ROLE))
            result = conn.execute_ddl("SHOW GRANT ROLE %s" % TEST_ROLE)
            self.assertEqual(len(result), 1)
            self.assertEqual(len(result[0]), 16)
            self.assertTrue(result[0][13])

           # Try to drop above restrict with a mismatch in transform function i.e
           # tokenize() instead of mask(), it should throw an exception
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.execute_ddl(
                    "DROP RESTRICT SELECT ON DATABASE %s %s WHERE %s FROM ROLE %s" % \
                    (TEST_DB, "TRANSFORM deny_test.v1 WITH tokenize()", "true", TEST_ROLE))
            self.assertTrue("CatalogException: Grant WITH TRANSFORM 'deny_test.v1:tokenize(__COLUMN__)' WITH FILTER 'TRUE' does not exist."
                            in str(ex_ctx.exception), msg=str(ex_ctx.exception))

            # Now try to drop it with no mismatch and it should be successful.
            conn.execute_ddl(
                "DROP RESTRICT SELECT ON DATABASE %s %s WHERE %s FROM ROLE %s" % \
                (TEST_DB, "TRANSFORM deny_test.v1 WITH mask()", "true", TEST_ROLE))
            result = conn.execute_ddl("SHOW GRANT ROLE %s" % TEST_ROLE)
            self.assertTrue(not result)

            #### Case 3: ####
            ## To verify that the entry for additional properties (like name, description)
            ## gets deleted from the kv_store table after DROP RESTRICT command

            # Create a Filter restrict
            props = " POLICYPROPERTIES('name'='test-drop-restrict', 'description'='policy to test drop restrict.')"
            conn.execute_ddl(
                "RESTRICT SELECT ON DATABASE %s WHERE %s FROM ROLE %s %s" % \
                (TEST_DB, "true", TEST_ROLE, props))
            result = conn.execute_ddl("SHOW GRANT ROLE %s" % TEST_ROLE)
            self.assertEqual(len(result), 1)
            self.assertEqual(len(result[0]), 16)
            self.assertTrue(result[0][13])

            # Map the OKEAR_DB.KV_STORE table to okera_kv_store table so that we can
            # run the query against it for verification
            conn.execute_ddl(
                """CREATE EXTERNAL TABLE %s.okera_kv_store STORED AS JDBC TBLPROPERTIES(
                'credentials.file' = '/opt/okera/system_tables.conf',
                'jdbc.schema.name'='public', 'table'='kv_store')""" % TEST_DB)

            # Verify that an entry for additional properties is created in KV_STORE table.
            like_clause = "'%test-drop-restrict%'"
            value_to_verify = """{"name":"test-drop-restrict","description":"policy to test drop restrict."}"""
            data = conn.scan_as_json(
                "SELECT * FROM %s.okera_kv_store WHERE value like %s" % (TEST_DB, like_clause))
            self.assertEqual(len(data), 1)
            self.assertTrue(value_to_verify in str(data))

            # Now, DROP RESTRICT priv
            conn.execute_ddl(
                "DROP RESTRICT SELECT ON DATABASE %s WHERE %s FROM ROLE %s %s" % \
                (TEST_DB, "true", TEST_ROLE, props))
            result = conn.execute_ddl("SHOW GRANT ROLE %s" % TEST_ROLE)
            self.assertTrue(not result)

            # Verify that the entry in the table KV_STORE should also get deleted.
            data = conn.scan_as_json(
                "SELECT * FROM %s.okera_kv_store WHERE value like %s" % (TEST_DB, like_clause))
            self.assertTrue(not data)

    def test_grant_props(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__cleanup(conn)
            props = " POLICYPROPERTIES('name'='me', 'description'='awesome!')"
            conn.execute_ddl("GRANT SELECT ON CATALOG TO ROLE %s %s" % (TEST_ROLE, props))

            params = _thrift_api.TListRolesParams()
            params.role_name_substring = TEST_ROLE
            params.include_privilege_props = True
            roles = conn._underlying_client().ListRoles(params)
            self.assertEqual(len(roles.roles), 1)
            role = roles.roles[0]
            self.assertEqual(len(role.privileges), 1)
            priv = role.privileges[0]
            self.assertEqual(priv.privilege_type, TPrivilegeType.GRANT)
            self.assertEqual(priv.privilege_name, 'server=server1->action=SELECT')
            # ID is auto-generated, it should be there
            self.assertTrue('id' in priv.props)
            id1 = priv.props['id']
            del priv.props['id']
            print(id1)
            self.assertEqual(priv.props, {'name':'me', 'description':'awesome!'})

            # check that getRole API also returns the props
            params = _thrift_api.TGetRoleParams()
            params.role_name = TEST_ROLE
            params.include_privilege_props = True
            role = conn._underlying_client().GetRole(params).role
            self.assertEqual(len(role.privileges), 1)
            priv = role.privileges[0]
            # ID is auto-generated, it should be there
            self.assertEqual(priv.privilege_type, TPrivilegeType.GRANT)
            self.assertTrue('id' in priv.props)
            id2 = priv.props['id']
            del priv.props['id']
            self.assertEqual(priv.privilege_name, 'server=server1->action=SELECT')
            self.assertEqual(priv.props, {'name':'me', 'description':'awesome!'})

            # IDs should match
            self.assertEqual(id1, id2)

            # Another grant, generate new ID
            props = " POLICYPROPERTIES('name'='me2', 'description'='awesome!')"
            conn.execute_ddl("GRANT SELECT ON DATABASE okera_sample TO ROLE %s %s" %\
                (TEST_ROLE, props))
            params = _thrift_api.TListRolesParams()
            params.role_name_substring = TEST_ROLE
            params.include_privilege_props = True
            roles = conn._underlying_client().ListRoles(params)
            self.assertEqual(len(roles.roles), 1)
            role = roles.roles[0]
            self.assertEqual(len(role.privileges), 2)
            priv = role.privileges[0]
            self.assertEqual(priv.privilege_type, TPrivilegeType.GRANT)
            self.assertEqual(priv.privilege_name,
                             'server=server1->db=okera_sample->action=SELECT')
            # ID is auto-generated, it should be there
            self.assertTrue('id' in priv.props)
            id3 = priv.props['id']
            del priv.props['id']
            self.assertEqual(priv.props, {'name':'me2', 'description':'awesome!'})
            self.assertTrue(id1 != id3)

            priv = role.privileges[1]
            self.assertEqual(priv.privilege_type, TPrivilegeType.GRANT)
            self.assertEqual(priv.privilege_name, 'server=server1->action=SELECT')
            # ID is auto-generated, it should be there
            self.assertTrue('id' in priv.props)
            id4 = priv.props['id']
            del priv.props['id']
            self.assertEqual(priv.props, {'name':'me', 'description':'awesome!'})
            self.assertEqual(id1, id4)

            # Issue an alter
            props = " POLICYPROPERTIES('name'='me3', 'id'='%s')" % id4
            conn.execute_ddl("ALTER %s" % props)

            roles = conn._underlying_client().ListRoles(params)
            self.assertEqual(len(roles.roles), 1)
            role = roles.roles[0]
            self.assertEqual(len(role.privileges), 2)
            priv = role.privileges[1]
            self.assertEqual(priv.privilege_type, TPrivilegeType.GRANT)

            # Should get same ID but different value
            self.assertTrue('id' in priv.props)
            id5 = priv.props['id']
            del priv.props['id']
            self.assertEqual(priv.props, {'name':'me3'})
            self.assertEqual(id4, id5)

    def test_restrict_filter(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__cleanup(conn)

            conn.execute_ddl(
                "RESTRICT SELECT ON DATABASE %s WHERE %s FROM ROLE %s" % \
                (TEST_DB, "int_col = 2", TEST_ROLE))
            conn.execute_ddl(
                "RESTRICT SELECT ON DATABASE %s WHERE %s FROM ROLE %s" % \
                (TEST_DB, "not_a_col = 1", TEST_ROLE))
            data = conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertEqual(len(data), 2)

            # Restricts only, no access granted
            ctx.enable_token_auth(token_str=TEST_USER)
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertTrue('does not have privilege' in str(ex_ctx.exception),
                            msg=str(ex_ctx.exception))

            ctx.disable_auth()
            conn.execute_ddl(
                "GRANT SELECT ON DATABASE %s TO ROLE %s" % \
                (TEST_DB, TEST_ROLE))

            ctx.enable_token_auth(token_str=TEST_USER)
            data = conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertEqual(len(data), 1)
            self.assertEqual(data[0]['int_col'], 2)

    def test_restrict_redact(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__cleanup(conn)

            conn.execute_ddl(
                "RESTRICT SELECT ON DATABASE %s HAVING ATTRIBUTE %s FROM ROLE %s" % \
                (TEST_DB, 'deny_test.v1', TEST_ROLE))
            data = conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertEqual(len(data), 2)
            self.assertEqual(data[0]['int_col'], 2)
            self.assertEqual(data[0]['smallint_col'], 1)
            self.assertEqual(data[1]['smallint_col'], 7)

            # Restricts only, no access granted
            ctx.enable_token_auth(token_str=TEST_USER)
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertTrue('does not have privilege' in str(ex_ctx.exception),
                            msg=str(ex_ctx.exception))

            ctx.disable_auth()
            conn.execute_ddl(
                "GRANT SELECT ON DATABASE %s TO ROLE %s" % \
                (TEST_DB, TEST_ROLE))

            ctx.enable_token_auth(token_str=TEST_USER)
            data = conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertEqual(len(data), 2)
            print(data)
            self.assertTrue('int_col' not in data[0])
            self.assertTrue('string_col' not in data[0])

    def test_restrict_transform(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__cleanup(conn)

            conn.execute_ddl(
                "RESTRICT SELECT ON DATABASE %s TRANSFORM %s WITH %s FROM ROLE %s" % \
                (TEST_DB, 'deny_test.v1', 'tokenize()', TEST_ROLE))
            data = conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertEqual(len(data), 2)
            self.assertEqual(data[0]['int_col'], 2)
            self.assertEqual(data[0]['smallint_col'], 1)
            self.assertEqual(data[1]['smallint_col'], 7)

            # Restricts only, no access granted
            ctx.enable_token_auth(token_str=TEST_USER)
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertTrue('does not have privilege' in str(ex_ctx.exception),
                            msg=str(ex_ctx.exception))

            ctx.disable_auth()
            conn.execute_ddl(
                "GRANT SELECT ON DATABASE %s TO ROLE %s" % \
                (TEST_DB, TEST_ROLE))

            ctx.enable_token_auth(token_str=TEST_USER)
            data = conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertEqual(len(data), 2)
            self.assertEqual(data[0]['int_col'], 733288971)
            self.assertEqual(data[1]['int_col'], 953486778)
            self.assertEqual(data[0]['smallint_col'], 1)
            self.assertEqual(data[1]['smallint_col'], 7)

    def test_demo_1(self):
        ctx = common.get_test_context()
        NOT_VISIBLE_DB = 'not_visible_db'
        with common.get_planner(ctx) as conn:
            self.__cleanup(conn)
            conn.execute_ddl("CREATE DATABASE IF NOT EXISTS   %s" % NOT_VISIBLE_DB)

            # Policy grant with tokenize
            conn.execute_ddl(
                "GRANT SELECT ON DATABASE %s EXCEPT TRANSFORM %s WITH %s TO ROLE %s" % \
                (TEST_DB, 'deny_test.s1', 'tokenize()', TEST_ROLE))
            self.assertTrue(NOT_VISIBLE_DB in conn.list_databases())
            self.assertTrue(TEST_DB in conn.list_databases())

            # Should tokenize
            ctx.enable_token_auth(token_str=TEST_USER)
            data = conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertEqual(len(data), 2)
            self.assertEqual(data[0]['string_col'], 'yexyi')
            self.assertEqual(data[1]['string_col'], 'gtyjx')

            # Database should not be accessible to this user.
            self.assertTrue(NOT_VISIBLE_DB not in conn.list_databases())
            self.assertTrue(TEST_DB in conn.list_databases())

            # Safeguard policy with mask should take precedence
            ctx.disable_auth()
            conn.execute_ddl(
                "RESTRICT SELECT ON CATALOG TRANSFORM %s WITH %s FROM ROLE %s" % \
                ('deny_test.s1', 'mask()', TEST_ROLE))

            # Should mask
            ctx.enable_token_auth(token_str=TEST_USER)
            data = conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertEqual(len(data), 2)
            self.assertEqual(data[0]['string_col'], 'XXXXX')
            self.assertEqual(data[1]['string_col'], 'XXXXX')

            # Safeguard policy should not change it
            self.assertTrue(NOT_VISIBLE_DB not in conn.list_databases())
            self.assertTrue(TEST_DB in conn.list_databases())

    def test_demo_2(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__cleanup(conn)

            # Safeguard policy with mask should take precedence
            conn.execute_ddl(
                "RESTRICT SELECT ON CATALOG TRANSFORM %s WITH %s FROM ROLE %s" % \
                ('deny_test.s1', 'mask()', TEST_ROLE))

            # Add WHERE filter as a grant
            conn.execute_ddl(
                "GRANT SELECT ON DATABASE %s WHERE %s TO ROLE %s" % \
                (TEST_DB, 'int_col = 2', TEST_ROLE))

            # Should mask and filter
            ctx.enable_token_auth(token_str=TEST_USER)
            data = conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertEqual(len(data), 1)
            self.assertEqual(data[0]['string_col'], 'XXXXX')

    def test_demo_3(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__cleanup(conn)

            # Add WHERE filter as a grant
            conn.execute_ddl(
                "GRANT SELECT ON DATABASE %s WHERE %s TO ROLE %s" % \
                (TEST_DB, 'true', TEST_ROLE))

            # Should show table
            ctx.enable_token_auth(token_str=TEST_USER)
            data = conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertEqual(len(data), 2)
            self.assertEqual(data[0]['string_col'], 'hello')
            tbls = conn.list_datasets(TEST_DB)
            self.assertEqual(len(tbls), 2)

    @pytest.mark.skip(reason="TDD")
    def test_filter_ignore_invalid(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__cleanup(conn)
            props = " POLICYPROPERTIES('row-filter.ignore_missing_columns'='true')"
            filter = "WHERE not_a_column = 1"
            conn.execute_ddl(
                "GRANT SELECT ON CATALOG %s TO ROLE %s %s" % (filter, TEST_ROLE, props))

            ctx.enable_token_auth(token_str=TEST_USER)
            data = conn.scan_as_json('%s.alltypes' % TEST_DB)
            print(data)

    @pytest.mark.skip(reason="TDD")
    def test_filter_invalid(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__cleanup(conn)
            props = " POLICYPROPERTIES('row-filter.ignore_missing_columns'='true')"
            filter = "WHERE not_a_column = 1"
            conn.execute_ddl(
                "GRANT SELECT ON CATALOG %s TO ROLE %s %s" % (filter, TEST_ROLE, props))

            ctx.enable_token_auth(token_str=TEST_USER)
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.scan_as_json('%s.alltypes' % TEST_DB)
            self.assertTrue('references a column' in str(ex_ctx.exception),
                            msg=str(ex_ctx.exception))

if __name__ == "__main__":
    unittest.main()




# Copyright 2018 Okera Inc.
from __future__ import absolute_import
import multiprocessing
import time
import os
import threading
import signal

def default_max_client_process_count():
  return multiprocessing.cpu_count() * 2

 # pylint: disable=too-many-instance-attributes
class ConcurrencyController():
  def __init__(self, worker_count=default_max_client_process_count()):
    self.pid = os.getpid()
    self.manager = multiprocessing.Manager()
    # pylint: disable=no-member,protected-access
    self.manager_pid = self.manager._process.pid
    self.task_queue = multiprocessing.JoinableQueue()
    self.input_queue = self.manager.Queue()
    self.output_queue = multiprocessing.Queue()
    self.errors_queue = self.manager.Queue()
    self.metrics_dict = self.manager.dict()
    self.worker_count = worker_count
    self.workers = []
    self.is_running = False

  def start(self):
    self.is_running = True
    self.workers = [TaskHandlerProcess(parent_pid=self.pid,
                                       manager_pid=self.manager_pid,
                                       task_queue=self.task_queue,
                                       output_queue=self.output_queue,
                                       metrics_dict=self.metrics_dict,
                                       errors_queue=self.errors_queue,
                                       concurrency_ctl=self)
                    for i in range(self.worker_count)]
    for worker in self.workers:
      worker.start()

  def stop(self):
    if self.is_running:
      self.is_running = False
      # pylint: disable=unused-variable
      for i in range(self.worker_count):
        self.enqueueTask(task=None)
      self.task_queue.close()
      self.task_queue.join()
      self.terminate()

  def terminate(self):
    for worker in self.workers:
      worker.terminate()
      worker.join()
    self.manager.shutdown()

  def enqueueTask(self, task):
    self.task_queue.put(task)

class TaskHandlerProcess(multiprocessing.Process):

  def __init__(self,
               parent_pid,
               manager_pid,
               task_queue,
               output_queue,
               metrics_dict,
               errors_queue,
               concurrency_ctl):

    multiprocessing.Process.__init__(self)
    self.should_exit = False
    self.parent_pid = parent_pid
    self.manager_pid = manager_pid
    self.task_queue = task_queue
    self.output_queue = output_queue
    self.metrics_dict = metrics_dict
    self.errors_queue = errors_queue
    self.concurrency_ctl = concurrency_ctl

  def watch_dog(self):
    while self.concurrency_ctl.is_running and not self.should_exit:
      self.check_parent_process()
      time.sleep(1)

  @classmethod
  def kill_proc(cls, pid):
    try:
      os.kill(pid, signal.SIGKILL)
    except ProcessLookupError:
      # We expect to hit this case when a worker tries to kill a manager process
      # that would have already been killed by another worker
      pass

  def check_parent_process(self):
    if self.parent_pid != os.getppid():
      print('Parent died! Terminating Worker Process: '
            'pid={} ppid={}'.format(os.getpid(), os.getppid()))
      # First attempt to terminate the multiprocessing.Manager process which is running
      # independently from the parent and worker processes.
      self.kill_proc(self.manager_pid)
      self.kill_proc(os.getpid())

  def run(self):
    watchdog_thread = threading.Thread(target=self.watch_dog, daemon=True)
    watchdog_thread.start()

    while self.concurrency_ctl.is_running:
      next_task = self.task_queue.get()
      if next_task is None:
        # Poison pill means shutdown
        self.task_queue.task_done()
        break
      next_task.set_handler(self)

      # We make an assumption here that the BaseBackgroundTask has a max_records
      # property which is not always true.  Therefore, we're squelching the AttributeError
      # exception if it occurs.
      try:
        next_task.max_records = self.metrics_dict.get('limit', None)
      except AttributeError:
        pass

      answer = next_task()
      if next_task.errors:
        for e in next_task.errors:
          self.errors_queue.put(e)
      self.output_queue.put(answer)
      self.task_queue.task_done()

    self.should_exit = True

class BaseBackgroundTask():
  def __init__(self, name):
    self._name = name
    self.handler = None

  def __call__(self):
    raise Exception('Invalid invocation of an abstract function: ' +
                    'BaseBackgroundTask::__call__')

  def __str__(self):
    return self.Name

  def set_handler(self, handler):
    self.handler = handler

  @property
  def Name(self):
    return  self._name

class OkeraWorkerException(BaseException):
  def __init__(self, original_error=None):
    # Call the base class constructor with the parameters it needs
    super(OkeraWorkerException, self).__init__()
    import traceback
    self.error = original_error
    self.callstack = ''.join(traceback.format_stack())
    if self.error:
      self.exception = ''.join(traceback.format_exception(etype=type(self.error),
                                                          value=self.error,
                                                          tb=self.error.__traceback__))

  def format_exception(self):
    return '{}\n{}'.format(self.callstack, self.exception)

  def print_exception(self):
    print(self.format_exception())



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for Presto in PyOkera
#
# pylint: disable=broad-except
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-locals
# pylint: disable=bad-continuation
# pylint: disable=broad-except

import os
from datetime import datetime

import pytest

import pytz
import requests
import urllib3
import prestodb

from okera import context
from okera.tests import pycerebro_test_common as common

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

DEFAULT_PRESTO_HOST = os.environ['ODAS_TEST_HOST']
DEFAULT_PRESTO_PORT = int(os.environ['ODAS_TEST_PORT_PRESTO_COORD_HTTPS'])
DEFAULT_PLANNER_PORT = int(os.environ['ODAS_TEST_PORT_PLANNER_THRIFT'])
BAD_PRESTO_PORT = 1234

@pytest.mark.filterwarnings('ignore::urllib3.exceptions.InsecureRequestWarning')
@pytest.mark.filterwarnings('ignore:numpy.dtype size changed')
class PrestoTest(common.TestBase):

    def _connect_presto_with_session_props(self, props):
        user = 'root'
        return prestodb.dbapi.connect(
            host=DEFAULT_PRESTO_HOST,
            port=DEFAULT_PRESTO_PORT,
            user=user,
            catalog='okera',
            http_scheme="https",
            auth=prestodb.auth.BasicAuthentication(user, user),
            isolation_level=prestodb.transaction.IsolationLevel.READ_UNCOMMITTED,
            session_properties=props
        )

    def _close_presto_connection(self, conn, cursor):
        if cursor:
            cursor.cancel()
        if conn:
            conn.close()

    def test_basic(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('select * from okera_sample.whoami',
                                        dialect='presto')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_auto_choose_dialect(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('select * from okera_sample.whoami')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_presto_disabled_auto_choose_dialect(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx) as conn:
            results = conn.scan_as_json('select * from okera_sample.whoami')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_presto_disabled_presto_dialect(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx) as conn:
            with pytest.raises(ValueError):
                conn.scan_as_json('select * from okera_sample.whoami',
                                  dialect='presto')

    def test_presto_enabled_okera_dialect(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('select * from okera_sample.whoami',
                                        dialect='okera')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_bad_presto_connection(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host=DEFAULT_PRESTO_HOST, port=DEFAULT_PLANNER_PORT,
                         presto_host=DEFAULT_PRESTO_HOST,
                         presto_port=BAD_PRESTO_PORT) as conn:
            with pytest.raises(requests.exceptions.ConnectionError):
                conn.scan_as_json('select * from okera_sample.whoami',
                                  dialect='presto')

    def test_presto_ctx_namespace(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('select * from okera_sample.whoami',
                                        dialect='presto')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_presto_ctx_bad_namespace(self):
        ctx = common.get_test_context(namespace='okera2')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            with pytest.raises(prestodb.exceptions.PrestoUserError):
                conn.scan_as_json('select * from okera_sample.whoami',
                                  dialect='presto')

    def test_presto_conn_namespace(self):
        ctx = common.get_test_context(namespace='okera_bad')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT,
                                namespace='okera') as conn:
            results = conn.scan_as_json('select * from okera_sample.whoami',
                                        dialect='presto')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_presto_conn_bad_namespace(self):
        ctx = common.get_test_context(namespace='okera_bad')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT,
                                namespace='okera2') as conn:
            with pytest.raises(prestodb.exceptions.PrestoUserError):
                conn.scan_as_json('select * from okera_sample.whoami',
                                  dialect='presto')

    def test_max_records(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json(
                'select * from okera_sample.users', dialect='presto', max_records=100)
            self.assertEqual(len(results), 100)
            results = conn.scan_as_pandas(
                'select * from okera_sample.users', dialect='presto', max_records=100)
            self.assertEqual(len(results), 100)

    def test_max_records_with_less_records(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('select * from okera_sample.users limit 50',
                                        dialect='presto', max_records=100)
            self.assertEqual(len(results), 50)
            results = conn.scan_as_pandas('select * from okera_sample.users limit 50',
                                          dialect='presto', max_records=100)
            self.assertEqual(len(results), 50)

    def test_json_pandas_equality(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            json_results = conn.scan_as_json('select * from okera_sample.users limit 50',
                                             dialect='presto', max_records=100)
            pandas_results = conn.scan_as_pandas(
                'select * from okera_sample.users limit 50',
                dialect='presto', max_records=100)
            self.assertEqual(json_results, pandas_results.to_dict('records'))

    def test_no_auth(self):
        ctx = context() # use default to ensure no auth
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            with pytest.raises(RuntimeError):
                conn.scan_as_json('select * from okera_sample.whoami',
                                  dialect='presto')

    def test_ctx_dialect(self):
        ctx = common.get_test_context(dialect='okera')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('okera_sample.whoami')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_ctx_dialect_override(self):
        ctx = common.get_test_context(dialect='presto')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('okera_sample.whoami', dialect='okera')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_complex_types(self):
        datasets = [
            'rs_complex.struct_t',
            'rs_complex.multiple_structs_nested',
            'rs_complex.map_t',
            'rs_complex.struct_nested',
            'rs_complex_parquet.map_struct_t',
            'rs_complex_parquet.map_struct_array_t',
        ]
        ctx = common.get_test_context(dialect='presto')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            for dataset in datasets:
                presto_results = conn.scan_as_json(
                    'select * from %s' % dataset, dialect='presto')
                okera_results = conn.scan_as_json(
                    'select * from %s' % dataset, dialect='okera', strings_as_utf8=True)
                self.assertEqual(okera_results, presto_results)

    def test_external_view(self):
        ctx = common.get_test_context(dialect='presto')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            conn.execute_ddl("create view if not exists default.external_view as select" +
                             " `role_name` from okera_system.role_names;")
            conn.execute_ddl("ALTER TABLE default.external_view SET TBLPROPERTIES" +
                             " ('cerebro.external.view'='true');")
            conn.execute_ddl("grant select on database default to okera_public_role;")

        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json("select * from default.external_view")
            self.assertTrue("role_name" in str(results), results)

        ctx.enable_token_auth(token_str='some_test_user')
        caught = None
        expected = "Unknown dataset: okera_system.role_names"
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            try:
                results = conn.scan_as_json("select * from default.external_view")
            except Exception as e:
                if expected not in str(e):
                    self.fail("Exception did not match expected. Ex encountered: "
                              + str(e))
                else:
                    caught = True
            if (not caught):
                self.fail("No exception raised, expected TRecordServiceException")

    def verify_allowed_view_auth_sql(self, user, sql):
        ctx = common.get_test_context(dialect='presto')
        ctx.enable_token_auth(token_str=user)
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            try:
                results = conn.scan_as_json(sql)
                self.assertTrue(results is not None)
            except Exception as e:
                self.fail("Fail to verify auth for user: " + user + " for sql: " + sql
                            + " due to Exception: " + str(e))

    def verify_restricted_view_auth_sql(self, user, expected, sql):
        ctx = common.get_test_context(dialect='presto')
        ctx.enable_token_auth(token_str=user)
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            caught = None
            try:
                results = conn.scan_as_json(sql)
                assert(results)
            except Exception as e:
                if expected not in str(e):
                    self.fail("Exception did not match expected. Ex encountered: "
                              + str(e))
                else:
                    caught = True
            if (not caught):
                self.fail(
                    "No exception raised, expected TRecordServiceException for user: " +
                    user + ", for sql: " + sql)

    def test_complex_type_view_auth(self):
        ## verify complex_type where clause auth
        where_qry = """select currency from chase.subscription_view
            where subscriptionkey = 'e23c82e3-10fd-40f3-ba59-5959a82ad3c8'"""
        ## verify complex_type Join clause auth
        join_query = """select a.currency from chase.subscription_view a
             join chase.subscription_view b
             ON (a.subscriptionkey = b.subscriptionkey)"""
        ## verify complex_type GROUP BY clause auth
        group_by_query = """select MAX(accountnumber), productkey
            from chase.subscription_view GROUP BY productkey"""
        ## verify complex_type UNION ALL clause auth
        union_all_query = """select count(*) from chase.subscription_view
            WHERE subscriptionkey='e23c82e3-10fd-40f3-ba59-5959a82ad3c8'
            UNION ALL select count(*) from chase.subscription_view
            WHERE subscriptionkey='e23c82e3-10fd-40f3-ba59-5959a82ad3c8'"""

        ## Run tests as 'root' user
        user = 'root'
        self.verify_allowed_view_auth_sql(user, where_qry)
        self.verify_allowed_view_auth_sql(user, join_query)
        self.verify_allowed_view_auth_sql(user, group_by_query)
        self.verify_allowed_view_auth_sql(user, union_all_query)

        ## Run some tests as 'testuser' user
        user = 'testuser'
        self.verify_allowed_view_auth_sql(user,
            "select * from chase.subscription_currency")
        self.verify_allowed_view_auth_sql(user,
            "select currency from chase.subscription_currency")
        self.verify_allowed_view_auth_sql(user,
            "select count(*) from chase.subscription_view " +
            "WHERE currency.currencycode='GBP'")
        allowed_union_all_query = """select count(*) from chase.subscription_view
            WHERE currency.currencycode='GBP'
            UNION ALL select count(*) from chase.subscription_view
            WHERE currency.currencycode='GBP'"""
        self.verify_allowed_view_auth_sql(user, allowed_union_all_query)

        self.verify_restricted_view_auth_sql(
            'testuser', "Column 'subscriptionkey' cannot be resolved", where_qry)
        self.verify_restricted_view_auth_sql(
            'testuser', "Column 'a.subscriptionkey' cannot be resolved", join_query)
        self.verify_restricted_view_auth_sql(
            'testuser', "Column 'accountnumber' cannot be resolved", group_by_query)
        self.verify_restricted_view_auth_sql(
            'testuser', "Column 'subscriptionkey' cannot be resolved", union_all_query)

    def test_auth_primitive_type_dataset(self):
        ## Regular SELECT queries on primitive datasets
        ## A "select_col" here means selective columns from the table/view
        select_queries = [
            """select * from authdb.alltypes_s3""",
            """select * from authdb.alltypes_s3_view""",
            """select * from authdb.alltypes_s3_sel_view""",
            """select * from authdb.alltypes_s3_view_on_view""",
            """select * from authdb.alltypes_s3_sel_view_on_sel_view""",
            """select * from authdb.alltypes_s3_full_view_on_sel_view""",
            """select * from authdb.alltypes_s3_where_clause_view""",
            """select * from authdb.alltypes_s3_join_clause_view""",
            """select int_col, string_col from authdb.alltypes_s3""",
            """select int_col from authdb.alltypes_s3_view""",
            """select int_col from authdb.alltypes_s3_sel_view""",
            """select int_col from authdb.alltypes_s3_view_on_view""",
            """select int_col from authdb.alltypes_s3_sel_view_on_sel_view""",
            """select int_col from authdb.alltypes_s3_full_view_on_sel_view""",
            """select int_col from authdb.alltypes_s3_where_clause_view""",
            """select int_col from authdb.alltypes_s3_join_clause_view""",
        ]

        ## Queries used to check column level restrictions
        ## These will only error out for 'testuser'/'noadmin' users,
        ## so DO NOT run for other users
        select_col_level_restrict_queries = [
            """select decimal_col from authdb.alltypes_s3""",
            """select decimal_col, double_col from authdb.alltypes_s3_view""",
            """select decimal_col from authdb.alltypes_s3_sel_view""",
            """select decimal_col from authdb.alltypes_s3_view_on_view""",
            """select decimal_col from authdb.alltypes_s3_sel_view_on_sel_view""",
            """select decimal_col from authdb.alltypes_s3_full_view_on_sel_view""",
            """select decimal_col from authdb.alltypes_s3_where_clause_view""",
            """select decimal_col from authdb.alltypes_s3_join_clause_view""",
        ]

        ## Run tests as 'root' user
        user = 'root'
        for query in select_queries:
            self.verify_allowed_view_auth_sql(user, query)

        ## Run tests as 'noadmin' user
        user = 'noadmin'
        expected = "User 'noadmin' does not have privileges to access: authdb.*"
        for query in select_queries:
            self.verify_restricted_view_auth_sql(user, expected, query)

        ## Run tests as 'selectall' user (allowed has SELECT ON CATALOG ACCESS)
        user = 'selectall'
        for query in select_queries:
            self.verify_allowed_view_auth_sql(user, query)

        ## Test for column level permissions
        ## Step 1: Run tests as 'testuser' user for allowed columns
        user = 'testuser'
        for query in select_queries:
            self.verify_allowed_view_auth_sql(user, query)

        ## Step 2: Run tests as 'testuser' user for restricted columns
        user = 'testuser'
        expected = "Column 'decimal_col' cannot be resolved"
        for query in select_col_level_restrict_queries:
            self.verify_restricted_view_auth_sql(user, expected, query)

    def test_auth_struct_type_dataset(self):
        ## Regular SELECT queries on struct_type datasets
        ## A "select_col" here means selective columns from the table/view
        select_queries = [
            """select * from authdb.struct_t""",
            """select * from authdb.struct_t_view""",
            """select * from authdb.struct_t_sel_view""",
            """select * from authdb.struct_t_view_on_view""",
            """select * from authdb.struct_t_sel_view_on_sel_view""",
            """select * from authdb.struct_t_full_view_on_sel_view""",
            """select * from authdb.struct_t_where_clause_view""",
            """select * from authdb.struct_t_join_clause_view""",
            """select id from authdb.struct_t""",
            """select id from authdb.struct_t_view""",
            """select id from authdb.struct_t_sel_view""",
            """select id from authdb.struct_t_view_on_view""",
            """select id from authdb.struct_t_sel_view_on_sel_view""",
            """select id from authdb.struct_t_full_view_on_sel_view""",
            """select id from authdb.struct_t_where_clause_view""",
            """select id from authdb.struct_t_join_clause_view""",
        ]

        ## Queries used to check column level restrictions
        ## These will only error out for 'testuser'/'noadmin' users
        select_col_level_restrict_queries = [
            """select s1 from authdb.struct_t""",
            """select s1 from authdb.struct_t_view""",
            """select s1 from authdb.struct_t_sel_view""",
            """select s1 from authdb.struct_t_view_on_view""",
            """select s1 from authdb.struct_t_sel_view_on_sel_view""",
            """select s1 from authdb.struct_t_full_view_on_sel_view""",
            """select s1 from authdb.struct_t_where_clause_view""",
            """select s1 from authdb.struct_t_join_clause_view""",
        ]

        ## Run tests as 'root' user
        user = 'root'
        for query in select_queries:
            self.verify_allowed_view_auth_sql(user, query)

        ## Run tests as 'noadmin' user
        user = 'noadmin'
        expected = "User 'noadmin' does not have privileges to access: authdb.*"
        for query in select_queries:
            self.verify_restricted_view_auth_sql(user, expected, query)

        ## Run tests as 'selectall' user (allowed has SELECT ON CATALOG ACCESS)
        user = 'selectall'
        for query in select_queries:
            self.verify_allowed_view_auth_sql(user, query)

        ## Test for column level permissions
        ## Step 1: Run tests as 'testuser' user for allowed columns
        user = 'testuser'
        for query in select_queries:
            self.verify_allowed_view_auth_sql(user, query)

        ## Step 2: Run tests as 'testuser' user for restricted columns
        user = 'testuser'
        expected = "Column 's1' cannot be resolved"
        for query in select_col_level_restrict_queries:
            self.verify_restricted_view_auth_sql(user, expected, query)

    def test_auth_struct_type_complex_type_access_dataset(self):
        ## Regular SELECT queries on struct_type datasets
        ## A "select_col" here means selective columns from the table/view
        select_queries = [
            """select * from authdb.struct_t""",
            """select * from authdb.struct_t_view""",
            """select * from authdb.struct_t_sel_complex_view""",
            """select * from authdb.struct_t_view_on_view""",
            """select * from authdb.struct_t_sel_view_on_sel_complex_view""",
            """select * from authdb.struct_t_full_view_on_sel_complex_view""",
            """select * from authdb.struct_t_where_clause_view""",
            """select * from authdb.struct_t_join_clause_view""",
            """select s1 from authdb.struct_t""",
            """select s1 from authdb.struct_t_view""",
            """select s1 from authdb.struct_t_sel_complex_view""",
            """select s1 from authdb.struct_t_view_on_view""",
            """select s1 from authdb.struct_t_sel_view_on_sel_complex_view""",
            """select s1 from authdb.struct_t_full_view_on_sel_complex_view""",
            """select s1 from authdb.struct_t_where_clause_view""",
            """select s1 from authdb.struct_t_join_clause_view""",
        ]

        ## Queries used to check column level restrictions
        ## These will only error out for 'complex' users
        select_col_level_restrict_queries = [
            """select id from authdb.struct_t""",
            """select id from authdb.struct_t_view""",
            """select id from authdb.struct_t_sel_complex_view""",
            """select id from authdb.struct_t_view_on_view""",
            """select id from authdb.struct_t_sel_view_on_sel_complex_view""",
            """select id from authdb.struct_t_full_view_on_sel_complex_view""",
            """select id from authdb.struct_t_where_clause_view""",
            """select id from authdb.struct_t_join_clause_view""",
        ]

        ## Test for column level permissions
        ## Step 1: Run tests as 'complex' user for allowed columns
        user = 'complex'
        for query in select_queries:
            self.verify_allowed_view_auth_sql(user, query)

        ## Step 2: Run tests as 'complex' user for restricted columns
        user = 'complex'
        expected = "Column 'id' cannot be resolved"
        for query in select_col_level_restrict_queries:
            self.verify_restricted_view_auth_sql(user, expected, query)


    def test_auth_map_type_dataset(self):
        ## Regular SELECT queries on map type datasets
        ## A "select_col" here means selective columns from the table/view
        select_queries = [
            """select * from authdb.map_t""",
            """select * from authdb.map_t_view""",
            """select * from authdb.map_t_sel_view""",
            """select * from authdb.map_t_view_on_view""",
            """select * from authdb.map_t_sel_view_on_sel_view""",
            """select * from authdb.map_t_full_view_on_sel_view""",
            """select * from authdb.map_t_where_clause_view""",
            """select * from authdb.map_t_join_clause_view""",
            """select str_map from authdb.map_t""",
            """select str_map from authdb.map_t_view""",
            """select str_map from authdb.map_t_sel_view""",
            """select str_map from authdb.map_t_view_on_view""",
            """select str_map from authdb.map_t_sel_view_on_sel_view""",
            """select str_map from authdb.map_t_full_view_on_sel_view""",
            """select str_map from authdb.map_t_where_clause_view""",
            """select str_map from authdb.map_t_join_clause_view""",
        ]

        ## Queries used to check column level restrictions
        ## These will only error out for 'testuser'/'noadmin' users,
        ## so DO NOT run for other users
        select_col_level_restrict_queries = [
            """select id from authdb.map_t""",
            """select id from authdb.map_t_view""",
            """select id from authdb.map_t_sel_view""",
            """select id from authdb.map_t_view_on_view""",
            """select id from authdb.map_t_sel_view_on_sel_view""",
            """select id from authdb.map_t_full_view_on_sel_view""",
            """select id from authdb.map_t_where_clause_view""",
            """select id from authdb.map_t_join_clause_view""",
        ]

        ## Run tests as 'root' user
        user = 'root'
        for query in select_queries:
            self.verify_allowed_view_auth_sql(user, query)

        ## Run tests as 'noadmin' user
        user = 'noadmin'
        expected = "User 'noadmin' does not have privileges to access: authdb.*"
        for query in select_queries:
            self.verify_restricted_view_auth_sql(user, expected, query)

        ## Run tests as 'selectall' user (allowed has SELECT ON CATALOG ACCESS)
        user = 'selectall'
        for query in select_queries:
            self.verify_allowed_view_auth_sql(user, query)

        ## Test for column level permissions
        ## Step 1: Run tests as 'testuser' user for allowed columns
        user = 'testuser'
        for query in select_queries:
            self.verify_allowed_view_auth_sql(user, query)

        ## Step 2: Run tests as 'testuser' user for restricted columns
        user = 'testuser'
        expected = "Column 'id' cannot be resolved"
        for query in select_col_level_restrict_queries:
            self.verify_restricted_view_auth_sql(user, expected, query)

    def test_auth_array_type_dataset(self):
        ## Regular SELECT queries on map type datasets
        ## A "select_col" here means selective columns from the table/view
        select_queries = [
            """select * from authdb.strarray_t""",
            """select * from authdb.strarray_t_view""",
            """select * from authdb.strarray_t_sel_view""",
            """select * from authdb.strarray_t_view_on_view""",
            """select * from authdb.strarray_t_sel_view_on_sel_view""",
            """select * from authdb.strarray_t_full_view_on_sel_view""",
            """select * from authdb.strarray_t_where_clause_view""",
            """select * from authdb.strarray_t_join_clause_view""",
            """select str_arr from authdb.strarray_t""",
            """select str_arr from authdb.strarray_t_view""",
            """select str_arr from authdb.strarray_t_sel_view""",
            """select str_arr from authdb.strarray_t_view_on_view""",
            """select str_arr from authdb.strarray_t_sel_view_on_sel_view""",
            """select str_arr from authdb.strarray_t_full_view_on_sel_view""",
            """select str_arr from authdb.strarray_t_where_clause_view""",
            """select str_arr from authdb.strarray_t_join_clause_view""",
        ]

        ## Queries used to check column level restrictions
        ## These will only error out for 'testuser'/'noadmin' users,
        ## so DO NOT run for other users
        select_col_level_restrict_queries = [
            """select id from authdb.strarray_t""",
            """select id from authdb.strarray_t_view""",
            """select id from authdb.strarray_t_sel_view""",
            """select id from authdb.strarray_t_view_on_view""",
            """select id from authdb.strarray_t_sel_view_on_sel_view""",
            """select id from authdb.strarray_t_full_view_on_sel_view""",
            """select id from authdb.strarray_t_where_clause_view""",
            """select id from authdb.strarray_t_join_clause_view""",
        ]

        ## Run tests as 'root' user
        user = 'root'
        for query in select_queries:
            self.verify_allowed_view_auth_sql(user, query)

        ## Run tests as 'noadmin' user
        user = 'noadmin'
        expected = "User 'noadmin' does not have privileges to access: authdb.*"
        for query in select_queries:
            self.verify_restricted_view_auth_sql(user, expected, query)

        ## Run tests as 'selectall' user (allowed has SELECT ON CATALOG ACCESS)
        user = 'selectall'
        for query in select_queries:
            self.verify_allowed_view_auth_sql(user, query)

        ## Test for column level permissions
        ## Step 1: Run tests as 'testuser' user for allowed columns
        user = 'testuser'
        for query in select_queries:
            self.verify_allowed_view_auth_sql(user, query)

        ## Step 2: Run tests as 'testuser' user for restricted columns
        user = 'testuser'
        expected = "Column 'id' cannot be resolved"
        for query in select_col_level_restrict_queries:
            self.verify_restricted_view_auth_sql(user, expected, query)

    def test_default_timezone(self):
        ctx = common.get_test_context(dialect='presto')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('select current_timezone() as tz')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['tz'], 'UTC')

    def test_no_timezone(self):
        ctx = common.get_test_context(dialect='presto')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('select current_timezone() as tz')
            self.assertEqual(len(results), 1)
            # Presto defaults to America/Los_Angeles assuming no configuration
            # has changed that. If that changes, this test will need changing
            # asw ell.
            self.assertEqual(results[0]['tz'], 'UTC')

    # pylint: disable=protected-access
    # is_std() is short for 'is standard time timezone'
    def is_std(self, timezone="UTC"):
        dt = datetime.utcnow()
        timezone = pytz.timezone(timezone)
        timezone_aware_date = timezone.localize(dt, is_dst=None)
        return timezone_aware_date.tzinfo._dst.seconds != 0
    # pylint: enable=protected-access

    def test_specific_timezones(self):
        tzs = [
            ('Asia/Calcutta', '+05:30'),
            ('America/New_York', '-04:00' if self.is_std('America/New_York')
                                          else '-05:00'),
            ('America/Los_Angeles', '-07:00' if self.is_std('America/Los_Angeles')
                                             else '-08:00'),
            ('Asia/Jerusalem', '+03:00' if self.is_std('Asia/Jerusalem')
                                        else '+02:00'),
        ]
        for tz in tzs:
            tz_name = tz[0]
            tz_offset = tz[1]
            ctx = common.get_test_context(dialect='presto', tz=pytz.timezone(tz_name))
            ctx.enable_token_auth(token_str='root')
            with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
                results = conn.scan_as_json('select current_timezone() as tz')
                self.assertEqual(len(results), 1)
                self.assertEqual(results[0]['tz'], tz_offset)

    # pylint: disable=protected-access
    def test_sample_and_limit_setting_via_session_props(self):
        # Verify that session properties allow to limit the out put of the presto queries.
        query_1_all_rows_length = 38455
        query_2_all_rows_length = 12
        sampling_value = 100
        expected_sampled_output_length = 3
        limit = 100
        test_query1 = 'select * from okera_sample.users'
        test_query2 = 'select * from partdb.ymdata2'
        test_query3 = 'select * from nytaxi.parquet_data'
        session_properties = {
            'okera.limit': limit,
            'okera.sampling_value': sampling_value
        }
        # first we set a connection without session properties
        # and check that we get all rows
        conn = self._connect_presto_with_session_props({})
        conn._http_session.verify = False
        cursor = conn.cursor()
        cursor.execute(test_query1)
        rows = cursor.fetchall()
        self.assertEqual(len(rows), query_1_all_rows_length)

        cursor.execute(test_query2)
        rows = cursor.fetchall()
        self.assertEqual(len(rows), query_2_all_rows_length)

        self._close_presto_connection(conn, cursor)

        # then we query the dataset that is essentially one file, so sampling_value
        # won't work but limit will
        # and check that output results were truncated according to limit
        conn = self._connect_presto_with_session_props(session_properties)
        conn._http_session.verify = False
        cursor = conn.cursor()
        cursor.execute(test_query1)
        rows = cursor.fetchall()
        self.assertEqual(len(rows), limit)

        # then we query the partitioned dataset so now the sampling_value will be
        # used rather than the limit.
        # It should restrict the output per file read to number of bytes specified
        cursor.execute(test_query2)
        rows = cursor.fetchall()
        self.assertEqual(len(rows), expected_sampled_output_length)
        self._close_presto_connection(conn, cursor)

        # then we query the big dataset to test both props
        # the query would hang normally without limit
        # the query should return sampled result
        with_limit_and_sampling_row_length = 700

        conn = self._connect_presto_with_session_props(session_properties)
        conn._http_session.verify = False
        cursor = conn.cursor()

        cursor.execute(test_query3)
        rows = cursor.fetchall()
        self.assertTrue(len(rows) <= with_limit_and_sampling_row_length)

        self._close_presto_connection(conn, cursor)

    # Only the first row in the result is verified.
    def verify_query(self, user, sql, expected_row):
        ctx = common.get_test_context(dialect='presto')
        ctx.enable_token_auth(token_str=user)
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            try:
                results = conn.scan_as_json(sql)
                self.assertTrue(results is not None)
                self.assertEqual(results, expected_row)
            except Exception as e:
                self.fail("Fail to verify results for user: " + user + " for sql: " + sql
                            + " due to Exception: " + str(e))

    def test_show_schemas(self):
        # verify database exists as root
        ctx = common.get_test_context(dialect='presto')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx) as conn:
            results = conn.execute_ddl("show databases like 'rs_complex'")
        expected = [["rs_complex"]]
        self.assertEqual(results, expected)

        # verify database is NOT found as 'analyst'
        ctx.enable_token_auth(token_str='analyst')
        with common.get_planner(ctx) as conn:
            results = conn.execute_ddl("show databases like 'rs_complex'")
        expected = [["rs_complex"]]
        self.assertNotEqual(results, expected)

    def test_transform(self):
        # verify query as analyst
        ctx = common.get_test_context(dialect='presto')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx) as conn:
            # grant privilege
            conn.execute_ddl("DROP ROLE IF EXISTS presto_view_role")
            conn.execute_ddl("CREATE ROLE presto_view_role")
            conn.execute_ddl("GRANT ROLE presto_view_role TO GROUP analyst")
            conn.execute_ddl("GRANT SELECT ON DATABASE abac_db " +
                             "HAVING attribute in " +
                             "(pii.ip_address, pii.email_address, pii.credit_card) " +
                             "TRANSFORM pii.email_address WITH tokenize() " +
                             "TRANSFORM pii.credit_card WITH mask_ccn() " +
                             "TO ROLE presto_view_role")
        user = 'analyst'
        sql = "SELECT * FROM okera.abac_db.user_account_data " +\
              "WHERE ipv4_address = '82.229.123.72'"

        expected_row = [{'ipv6_address': '7553:abe3:ca2f:9ac3:42b5:4f3b:7a46:9d52',
                     'email': 'Vafcw@yrpq.ryi',
                     'ipv4_address': '82.229.123.72',
                     'creditcardnumber': 'XXXXXXXXXXXXXXX3020'}]
        self.verify_query(user, sql, expected_row)

        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx) as conn:
            # revoke privilege
            conn.execute_ddl("REVOKE SELECT ON DATABASE abac_db " +
                             "HAVING attribute in " +
                             "(pii.ip_address, pii.email_address, pii.credit_card) " +
                             "TRANSFORM pii.email_address WITH tokenize() " +
                             "TRANSFORM pii.credit_card WITH mask_ccn() " +
                             "FROM ROLE presto_view_role")

    def test_decimal(self):
        sql = "SELECT sum(cs_sales_price) FROM okera.tpcds1_unpartitioned.catalog_sales"
        expected_row = [{'_col0': '72371322.53'}]
        self.verify_query('root', sql, expected_row)

    def test_cte_view(self):
        # verify external view with non-ODAS sql
        # executed by analyst user in presto
        ctx = common.get_test_context(dialect='presto')
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx) as conn:
            # grant privilege
            conn.execute_ddl("DROP ROLE IF EXISTS presto_view_role")
            conn.execute_ddl("CREATE ROLE presto_view_role")
            conn.execute_ddl("GRANT ROLE presto_view_role TO GROUP analyst")
            conn.execute_ddl("DROP ATTRIBUTE IF EXISTS chase_pii.money")
            conn.execute_ddl("CREATE ATTRIBUTE chase_pii.money")
            conn.execute_ddl("ALTER TABLE chase.ledger_balance " +\
                             "ADD COLUMN ATTRIBUTE " +\
                             "posted_balance chase_pii.money")
            conn.execute_ddl("ALTER TABLE chase.ledger_balance " +\
                             "ADD COLUMN ATTRIBUTE " +\
                             "available_balance chase_pii.money")
            conn.execute_ddl("GRANT SELECT ON DATABASE chase " +\
                             "TRANSFORM chase_pii.money WITH tokenize() " +\
                             "TO ROLE presto_view_role")

        user = 'analyst'
        sql = "select * from okera.chase.cte_view " +\
              "where dt = cast('2020-01-02 14:29:27.145' as timestamp)"

        expected_row = [{'dt': '2020-01-02 14:29:27.145',
                         'sum_posted_balance': '3408793.163',
                         'metric': 'Value credit transaction',
                         'sum_avbl_balance': '3408793.163'}]
        self.verify_query(user, sql, expected_row)

        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx) as conn:
            # revoke privilege
            conn.execute_ddl("REVOKE SELECT ON DATABASE chase " +\
                             "TRANSFORM chase_pii.money WITH tokenize() " +\
                             "FROM ROLE presto_view_role")
            conn.execute_ddl("ALTER TABLE chase.ledger_balance " +\
                             "DROP COLUMN ATTRIBUTE " +\
                             "available_balance chase_pii.money")
            conn.execute_ddl("ALTER TABLE chase.ledger_balance " +\
                             "DROP COLUMN ATTRIBUTE " +\
                             "posted_balance chase_pii.money")
            conn.execute_ddl("DROP ATTRIBUTE IF EXISTS chase_pii.money")

    def test_create_view(self):
        if common.TEST_LEVEL in ["dev", "smoke"]:
            # Run over fewer dbs in lower test modes
            dbs = [
                'rs_json_format',
            ]
        else:
            dbs = [
                'rs',
                'rs_complex',
                'rs_complex_parquet',
                'rs_json_format',
            ]
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            DB = 'test_presto_views'

            # Cleanup from previous runs
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % DB)
            conn.execute_ddl('CREATE DATABASE %s' % DB)

            skippable_datasets = [
                # Presto doesn't like views that contain keywords or
                # start with numbers
                'null',
                '12345_begins_with_number',
                '12345_begins_with_number2',
                # These files are too big to be useful, scan takes too long
                'alltypes_large_s3',
                'alltypes_large_s3_gz',
                'alltypes_large_s3_gz_no_ext',
                'alltypes_large_s3_lzo',
                'gharch_test_data',
                # Fails to parse the AVSC file
                'avro_comments',
                # Fails due to map in arrays
                'complextypestbl_parq',
                # TODO: there seems to be some utf-8 decoding issue
                'encoding',
                # presto and okera don't order this the same
                'increasing_file_sizes',
                'partitioned_alltypes',
                'partitioned_tbl_s3_from_readonly_bucket',
                'partitioned_tbl',
                'market_decide_single_avro',
                'spark_snappy_part',
                'partitioned_data',
                'zd1318_chase_1',
                'zd1318_chase_2',
                'zd1318_okera',
                'chase_ledger_balance_timestamp',
                'chase_ledger_transaction_timestamp',
                'chase_parquet_timestamp',
                # ODAS can't run this either
                'partitioned_tbl_s3',
                's3_no_perm',
                'array_text',
                'map_text',
                'rs_complex_array_map_t',
                'rs_parquet_array_map_t',
                # Table is too big for Presto to run
                'test_many_columns',
                # Is an external view that Presto can't parse/analyze
                'zd1075_view',
                # These hudi tables do not return data (due to subdirs)
                'hudi_as_parquet',
                'hudi_nonpartitioned',
            ]

            # we have some ordering issues when comparing results on some datasets.
            compare_records_returned_only = [
              'hudi_partitioned',
              'lineitem_orc',
            ]

            for db in dbs:
                datasets = conn.list_dataset_names(db)
                for dataset in datasets:
                    view_db, view_ds = dataset.split(".")
                    view_name = dataset.replace('.', '_')

                    # Skip datasets that don't work
                    if view_ds.lower() in skippable_datasets:
                        continue

                    print("Running test over %s.%s" % (view_db, view_ds))

                    # This is called 'scan' but is really just a way to execute a query
                    # against Presto, which does not distinguish DDL and DQL
                    create_view_stmt = 'CREATE VIEW %s.%s AS SELECT * FROM "%s"."%s"' % (
                        DB,
                        view_name,
                        view_db,
                        view_ds
                    )
                    res = conn.scan_as_json(create_view_stmt, dialect='presto')
                    assert len(res) == 1
                    assert 'result' in res[0] and res[0]['result']

                    presto_view_results = conn.scan_as_json(
                        'select * from %s.%s' % (DB, view_name), dialect='presto')
                    presto_ds_results = conn.scan_as_json(
                        'select * from %s' % dataset, dialect='presto')
                    okera_ds_results = conn.scan_as_json(
                        'select * from %s' % dataset, dialect='okera',
                        strings_as_utf8=True)

                    if view_ds.lower() in compare_records_returned_only:
                        self.assertEqual(len(presto_view_results),
                                         len(presto_ds_results))
                        self.assertEqual(len(presto_view_results),
                                         len(okera_ds_results))
                        continue

                    PrestoTest.compare_json(
                        self, presto_view_results, okera_ds_results, False,
                        empty_struct_equals_null=False,
                        batch_mode=False, required_type=None,
                        empty_array_equals_null=False)
                    PrestoTest.compare_json(
                        self, presto_view_results, presto_ds_results, False,
                        empty_struct_equals_null=False,
                        batch_mode=False, required_type=None,
                        empty_array_equals_null=False)

            views = conn.list_dataset_names(db=DB)
            for view in views:
                drop_view_stmt = 'DROP VIEW %s' % view
                res = conn.scan_as_json(drop_view_stmt, dialect='presto')
                assert len(res) == 1
                assert 'result' in res[0] and res[0]['result']

    def test_create_view_permission(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            DB = 'test_presto_views'

            # Cleanup from previous runs
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % DB)
            conn.execute_ddl('CREATE DATABASE %s' % DB)

            # This is called 'scan' but is really just a way to execute a query
            # against Presto, which does not distinguish DDL and DQL
            ctx.enable_token_auth(token_str='user_without_priv')
            with pytest.raises(prestodb.exceptions.PrestoQueryError) as excinfo:
                create_view_stmt = 'CREATE VIEW %s.test_view AS SELECT 1 as a' % DB
                conn.scan_as_json(create_view_stmt, dialect='presto')
            assert "not have privileges to execute 'CREATE'" in str(excinfo.value)

    def test_session_properties(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            with pytest.raises(prestodb.exceptions.PrestoQueryError) as excinfo:
                conn.scan_as_json(
                    'select count(*) from nytaxi.parquet_data',
                    dialect='presto',
                    presto_session={'query_max_execution_time': '1s'})
            assert "name=EXCEEDED_TIME_LIMIT" in str(excinfo.value)

            with pytest.raises(prestodb.exceptions.PrestoQueryError) as excinfo:
                conn.scan_as_pandas(
                    'select count(*) from nytaxi.parquet_data',
                    dialect='presto',
                    presto_session={'query_max_execution_time': '1s'})
            assert "name=EXCEEDED_TIME_LIMIT" in str(excinfo.value)

    def test_table_stats(self):
        DB = "stats_test_db"
        TBL = "stats_tbl"
        TBL_DDL = "CREATE TABLE %s.%s (c1 int)" % (DB, TBL)

        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            conn.execute_ddl("DROP DATABASE IF EXISTS %s CASCADE" % DB)
            conn.execute_ddl("CREATE DATABASE %s" % DB)
            conn.execute_ddl(TBL_DDL)

            res = conn.scan_as_json('show stats for %s.%s' % (DB, TBL))
            assert len(res) == 2
            assert res[1]['row_count'] is None

            alter_ddl = "ALTER TABLE %s.%s SET TBLPROPERTIES('numRows'='27')" % (DB, TBL)
            # TODO: Not sure why we need this twice, but essentially it's not reliably
            # set in the planner (even if manually checked) if we only execute this once.
            res = conn.execute_ddl(alter_ddl)
            res = conn.execute_ddl(alter_ddl)

            res = conn.scan_as_json('show stats for %s.%s' % (DB, TBL))
            assert len(res) == 2
            assert res[1]['row_count'] == 27

    def test_presto_udfs(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            res = conn.scan_as_json('select current_database()')
            assert res[0]['_col0'] == 'okera'

    def test_presto_password(self):
        query = 'select * from okera_sample.sample'
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='presto_test_user')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            res = conn.scan_as_json(query, dialect='presto')
            assert len(res) == 2

            ctx.enable_token_auth(token_str='presto_test_user', presto_password='abc')
            with pytest.raises(prestodb.exceptions.DatabaseError) as excinfo:
                conn.scan_as_json(query, dialect='presto')
            assert "401" in str(excinfo.value)

            ctx.enable_token_auth(token_str='presto_test_user',
                                  presto_password='presto_test_user')
            res = conn.scan_as_json(query, dialect='presto')
            assert len(res) == 2

    def test_plan_error(self):
        DB = "presto_error_db"

        ctx = common.get_test_context()
        ctx.enable_token_auth(token_str='root')
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            conn.execute_ddl("DROP DATABASE IF EXISTS %s CASCADE" % DB)
            conn.execute_ddl("CREATE DATABASE %s" % DB)
            conn.execute_ddl("CREATE TABLE %s.tbl(c1 string)" % DB)
            conn.execute_ddl("CREATE VIEW %s.vw AS SELECT c1 FROM %s.tbl" % (DB, DB))
            conn.execute_ddl("ALTER TABLE %s.tbl REPLACE COLUMNS (c2 string)" % DB)

            with pytest.raises(prestodb.exceptions.PrestoQueryError) as excinfo:
                conn.scan_as_json('select * from %s.vw' % (DB), dialect='presto')
            assert "Could not resolve column/field reference: 'c1'" in str(excinfo.value)

    def test_keywords_in_column_names(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx, presto_port=DEFAULT_PRESTO_PORT) as conn:
            DB = "presto_test_db"
            TBL = "test_table"

            ddls = [
                "DROP DATABASE IF EXISTS %s CASCADE" % DB,
                "CREATE DATABASE %s" % DB,
                "CREATE TABLE %s.%s(`database` int, `metadata` struct<`table`: int>)" % (DB, TBL)
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            ctx.enable_token_auth(token_str='root')
            conn.scan_as_json('SELECT * FROM %s.%s' % (DB, TBL), dialect='presto')

    # pylint: enable=protected-access



# Copyright 2020 Okera Inc. All Rights Reserved.
#
# Some integration tests for auth in PyOkera
#
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=no-else-return
# pylint: disable=duplicate-code

import unittest
import os

#from okera import context, _thrift_api
#from datetime import datetime
from okera.tests import pycerebro_test_common as common
from okera._thrift_api import (TAttribute, TAttributeMatchLevel, TAccessPermissionLevel)

TEST_USER = 'list_dbs_test_user'
TEST_ROLE = 'list_dbs_test_role'

TEST_MODE_GLUE = False
# We should do better check here to assert for True.
if 'TEST_MODE_GLUE' in os.environ:
    TEST_MODE_GLUE = True

class ListDatabasesTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """ Initializes one time state that is shared across test cases. This is used
            to speed up the tests. State that can be shared across (but still stable)
            should be here instead of __cleanup().
        """
        super(ListDatabasesTest, cls).setUpClass()

    def _get_t_attribute_obj(self, namespace, key):
        attribute = TAttribute()
        attribute.attribute_namespace = namespace
        attribute.key = key
        return attribute

    def test_list_databases_basic(self):
        USER = 'mktg_analyst'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl("create role if not exists temp_role")
            conn.execute_ddl("GRANT ROLE temp_role TO GROUP temp_user")
            conn.execute_ddl("create database if not exists temp_database")
            conn.execute_ddl("""
                grant CREATE_AS_OWNER on database temp_database to role temp_role""")
            # list databases test (exact match)
            result = conn.list_databases_v2(exact_names_filter=['demo_test'],
                                            limit=10, page_token=None)
            self.assertTrue(len(result.databases) == 1)

            # list databases test (pattern match)
            result = conn.list_databases_v2(name_pattern_filter='*tpc*',
                                            limit=10, page_token=None)
            self.assertTrue(len(result.databases) == 10)

            result = conn.list_databases_v2(name_pattern_filter='okera_sample|rs',
                                            limit=10, page_token=None)
            self.assertTrue(len(result.databases) == 2)

            # list databases test (no filter criteria)
            result = conn.list_databases_v2()
            self.assertTrue(len(result.databases) == 10)

            # list databases test (very weird pattern. Should return empty list)
            result = conn.list_databases_v2(name_pattern_filter='*WeirdPattern*',
                                            limit=10, page_token=None)
            self.assertTrue(len(result.databases) == 0)

            # list databases test (for access levels)
            ctx.enable_token_auth(token_str=USER)
            result = conn.list_databases_v2(exact_names_filter=['marketing_database'],
                                            limit=10, page_token=None)
            for database in result.databases:
                self.assertTrue(database.name == ['marketing_database'])
                self.assertTrue(database.datasets_count == 3)
                assert TAccessPermissionLevel.SHOW in database.access_levels
                assert TAccessPermissionLevel.CREATE_AS_OWNER in database.access_levels
                assert TAccessPermissionLevel.SELECT not in database.access_levels
            self.assertTrue(len(result.databases) == 1)

            # list databases test (when there are no tables in the db)
            ctx.enable_token_auth(token_str='temp_user')
            result = conn.list_databases_v2(exact_names_filter=['temp_database'],
                                            limit=10, page_token=None)
            for database in result.databases:
                self.assertTrue(database.name == ['temp_database'])
                self.assertTrue(database.datasets_count == 0)
                assert TAccessPermissionLevel.SHOW not in database.access_levels
                assert TAccessPermissionLevel.CREATE_AS_OWNER in database.access_levels
                assert TAccessPermissionLevel.SELECT not in database.access_levels
            self.assertTrue(len(result.databases) == 1)

            # Drop temp_database
            ctx.enable_token_auth(token_str='root')
            conn.execute_ddl('DROP DATABASE IF EXISTS temp_database CASCADE')

    @unittest.skipIf(not TEST_MODE_GLUE,
                     "Estimated count only supported for glue.")
    def test_list_db_estimated_count(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # list databases test (no filter criteria)
            result = conn.list_databases_v2(name_pattern_filter='okera*', limit=1)
            print(result)
            self.assertTrue(result is not None)
            self.assertTrue(result.databases[0].is_datasets_count_estimated is False)

            # create databases
            conn.execute_ddl('DROP DATABASE IF EXISTS page_test_db CASCADE')
            conn.execute_ddl('CREATE DATABASE page_test_db')

            # create 101 tables 0-100
            for i in range(101):
                tbl = "t_%s" % i
                conn.execute_ddl('CREATE TABLE page_test_db.%s(c1 int, c11 int)' % tbl)

            result = conn.list_databases_v2(exact_names_filter=['page_test_db'])
            print(result)
            self.assertTrue(result is not None)
            self.assertTrue(result.databases[0].datasets_count == 100)
            self.assertTrue(result.databases[0].is_datasets_count_estimated is True)

            # drop and re-create test role and assign it to a group
            # it would lose its old privileges.
            conn.execute_ddl('DROP ROLE IF EXISTS %s'% TEST_ROLE)
            conn.execute_ddl('CREATE ROLE %s'% TEST_ROLE)
            conn.execute_ddl('GRANT ROLE %s TO GROUP %s' % (TEST_ROLE, TEST_USER))
            # GRANT SELECT on just one table for the role TEST_ROLE.
            conn.execute_ddl('GRANT SELECT ON TABLE %s to ROLE %s' %
                             ("page_test_db.t_1", TEST_ROLE))

            # list databases test for user who has only one table access.
            # the flag is_datasets_count_estimated should be False in this case.
            ctx.enable_token_auth(token_str=TEST_USER)
            result = conn.list_databases_v2(exact_names_filter=['page_test_db'])
            print(result)
            self.assertTrue(result is not None)
            self.assertTrue(result.databases[0].datasets_count == 1)
            self.assertTrue(result.databases[0].is_datasets_count_estimated is False)

            ctx.enable_token_auth(token_str='root')
            conn.execute_ddl('DROP ROLE IF EXISTS %s' % TEST_ROLE)
            conn.execute_ddl('DROP DATABASE IF EXISTS page_test_db CASCADE')


    def test_list_databases_tag_filter1(self):
        """ Test to verify that list_databases() api returns correct databases after
            applying attributes (tag) filter with combination of any level of match
            i.e. DATABASE_ONLY, DATABASE_PLUS, TABLE_ONLY, COLUMN_ONLY etc.
            Scenario to test is different tags on the same database  at different levels.
        """
        db1 = 'attributes_test_db1'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # create attributes
            conn.delete_attribute('test_namespace', 'test_key1')
            conn.create_attribute('test_namespace', 'test_key1')
            conn.delete_attribute('test_namespace', 'test_key2')
            conn.create_attribute('test_namespace', 'test_key2')
            conn.delete_attribute('test_namespace', 'test_key3')
            conn.create_attribute('test_namespace', 'test_key3')

            # create databases
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db1)
            conn.execute_ddl('CREATE DATABASE %s' % db1)

            # create tables
            conn.execute_ddl('CREATE TABLE %s.t1(c1 int, c11 int)' % db1)

            # Assign attributes at DB level, Table level and Column level
            conn.assign_attribute('test_namespace', 'test_key1', db1)
            conn.assign_attribute('test_namespace', 'test_key2', db1, 't1')
            conn.assign_attribute('test_namespace', 'test_key3', db1, 't1', 'c1')

            # Create TAttribute objects required for testing
            attributes = []
            attr1_db_level = self._get_t_attribute_obj('test_namespace', 'test_key1')
            attr1_tab_level = self._get_t_attribute_obj('test_namespace', 'test_key2')
            attr1_col_level = self._get_t_attribute_obj('test_namespace', 'test_key3')

            # should return the database: db1
            # Input tag is on DATABASE level and Match_Level = DATABASE_ONLY
            # Tag filter matches at DB level only
            attributes.append(attr1_db_level)
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.DATABASE_ONLY)
            self.assertTrue(len(result.databases) == 1)
            for database in result.databases:
                self.assertTrue(database.name == [db1])

            # should return the database: db1
            # Input tag is on DATABASE level and Match_Level = DATABASE_PLUS
            # Tag filter matches at DB, Table and Column level
            attributes.clear()
            attributes.append(attr1_db_level)
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.DATABASE_PLUS)
            self.assertTrue(len(result.databases) == 1)
            for database in result.databases:
                self.assertTrue(database.name == [db1])

            # No Db should be returned.
            # Input tags are on TABLE and COLUMN level and Match_Level = DATABASE_ONLY
            # Tag filter matches at DB level only
            attributes.clear()
            attributes.append(attr1_tab_level)
            attributes.append(attr1_col_level)
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.DATABASE_ONLY)
            self.assertTrue(len(result.databases) == 0)

            # should return the database: db1
            # Input tags are on TABLE and COLUMN level and Match_Level = DATABASE_PLUS
            # Tag filter matches at DB, TABLE and COLUMN level
            attributes.clear()
            attributes.append(attr1_tab_level)
            attributes.append(attr1_col_level)
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.DATABASE_PLUS)
            self.assertTrue(len(result.databases) == 1)
            for database in result.databases:
                self.assertTrue(database.name == [db1])

            # drop databases
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db1)

    def test_list_databases_tag_filter2(self):
        """ Test to verify that list_databases() api returns correct databases after
            applying attributes (tag) filter with combination of any level of match
            i.e. DATABASE_ONLY, DATABASE_PLUS, TABLE_ONLY, COLUMN_ONLY etc.
            Scenario to test is different tags on different databases at different levels.
        """
        db2 = 'attributes_test_db2'
        db3 = 'attributes_test_db3'
        db4 = 'attributes_test_db4'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # create attributes
            conn.delete_attribute('test_namespace', 'test_key4')
            conn.create_attribute('test_namespace', 'test_key4')
            conn.delete_attribute('test_namespace', 'test_key5')
            conn.create_attribute('test_namespace', 'test_key5')
            conn.delete_attribute('test_namespace', 'test_key6')
            conn.create_attribute('test_namespace', 'test_key6')

            # create databases
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db2)
            conn.execute_ddl('CREATE DATABASE %s' % db2)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db3)
            conn.execute_ddl('CREATE DATABASE %s' % db3)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db4)
            conn.execute_ddl('CREATE DATABASE %s' % db4)

            # create tables
            conn.execute_ddl('CREATE TABLE %s.t2(c2 int, c22 int)' % db2)
            conn.execute_ddl('CREATE TABLE %s.t3(c3 int, c33 int)' % db3)
            conn.execute_ddl('CREATE TABLE %s.t4(c4 int, c44 int)' % db4)

            # Assign attributes at DB level, Table level and Column level
            conn.assign_attribute('test_namespace', 'test_key4', db2)
            conn.assign_attribute('test_namespace', 'test_key5', db3, 't3')
            conn.assign_attribute('test_namespace', 'test_key6', db4, 't4', 'c4')

            attributes = []
            attr2_db_level = self._get_t_attribute_obj('test_namespace', 'test_key4')
            attr3_tab_level = self._get_t_attribute_obj('test_namespace', 'test_key5')
            attr4_col_level = self._get_t_attribute_obj('test_namespace', 'test_key6')
            attributes.append(attr2_db_level)
            attributes.append(attr3_tab_level)
            attributes.append(attr4_col_level)

            # should return the database: db2
            # Input tags are on DB, TABLE, & COLUMN level and Match_Level = DATABASE_ONLY
            # Tag filter matches at DB level only
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.DATABASE_ONLY)
            self.assertTrue(len(result.databases) == 1)
            for database in result.databases:
                self.assertTrue(database.name == [db2])

            # should return the database: db3
            # Input tags are on DB, TABLE, & COLUMN level and Match_Level = TABLE_ONLY
            # Tag filter matches at TABLE level only
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.TABLE_ONLY)
            self.assertTrue(len(result.databases) == 1)
            for database in result.databases:
                self.assertTrue(database.name == [db3])

            # should return the database: db4
            # Input tags are on DB, TABLE, & COLUMN level and Match_Level = COLUMN_ONLY
            # Tag filter matches at COLUMN level only
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.COLUMN_ONLY)
            self.assertTrue(len(result.databases) == 1)
            for database in result.databases:
                self.assertTrue(database.name == [db4])

            # should return the database: db2, db3 and db4
            # Input tags are on DB, TABLE, & COLUMN level and Match_Level = DATABASE_PLUS
            # Tag filter matches at DB, TABLE, & COLUMN level
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.DATABASE_PLUS)
            self.assertTrue(len(result.databases) == 3)
            returned_dbs = []
            for database in result.databases:
                returned_dbs.append(database.name)
            self.assertTrue([db2] in returned_dbs)
            self.assertTrue([db3] in returned_dbs)
            self.assertTrue([db4] in returned_dbs)

            # should return the database: db3 and db4
            # Input tags are on DB, TABLE, & COLUMN level and Match_Level = TABLE_PLUS
            # Tag filter matches at TABLE & COLUMN level
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.TABLE_PLUS)
            self.assertTrue(len(result.databases) == 2)
            returned_dbs.clear()
            for database in result.databases:
                returned_dbs.append(database.name)
            self.assertTrue([db3] in returned_dbs)
            self.assertTrue([db4] in returned_dbs)

            # should return the database: db2, db3 and db4
            # Input tags are on DB, TABLE, & COLUMN level and Match_Level = None or NULL
            # if no Match_Level is specified, Tag filter matches at ALL levels
            result = conn.list_databases_v2(tags=attributes)
            self.assertTrue(len(result.databases) == 3)
            returned_dbs.clear()
            for database in result.databases:
                returned_dbs.append(database.name)
            self.assertTrue([db2] in returned_dbs)
            self.assertTrue([db3] in returned_dbs)
            self.assertTrue([db4] in returned_dbs)

            # drop databases
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db2)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db3)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db4)

    def test_list_databases_tag_filter3(self):
        """ Test to verify that list_databases() api returns correct databases after
            applying attributes (tag) filter with combination of any level of match
            i.e. DATABASE_ONLY, DATABASE_PLUS, TABLE_ONLY, COLUMN_ONLY etc.
            Scenario to test is same tag on different databases at different levels.
        """
        db5 = 'attributes_test_db5'
        db6 = 'attributes_test_db6'
        db7 = 'attributes_test_db7'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # create attribute
            conn.delete_attribute('test_namespace', 'test_key7')
            conn.create_attribute('test_namespace', 'test_key7')

            # create databases
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db5)
            conn.execute_ddl('CREATE DATABASE %s' % db5)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db6)
            conn.execute_ddl('CREATE DATABASE %s' % db6)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db7)
            conn.execute_ddl('CREATE DATABASE %s' % db7)

            # create tables
            conn.execute_ddl('CREATE TABLE %s.t5(c5 int, c55 int)' % db5)
            conn.execute_ddl('CREATE TABLE %s.t6(c6 int, c66 int)' % db6)
            conn.execute_ddl('CREATE TABLE %s.t7(c7 int, c77 int)' % db7)

            # assign same key (common key) to multiple dbs at different levels
            conn.assign_attribute('test_namespace', 'test_key7', db5)
            conn.assign_attribute('test_namespace', 'test_key7', db6, 't6')
            conn.assign_attribute('test_namespace', 'test_key7', db7, 't7', 'c7')

            attributes = []
            attr7_all_level = self._get_t_attribute_obj('test_namespace', 'test_key7')
            attributes.append(attr7_all_level)

            # should return the database: db5
            # Input tag is on DB, TABLE, & COLUMN level and Match_Level = DATABASE_ONLY
            # Tag filter matches at DATABASE level only
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.DATABASE_ONLY)
            self.assertTrue(len(result.databases) == 1)
            returned_dbs = []
            for database in result.databases:
                returned_dbs.append(database.name)
            self.assertTrue([db5] in returned_dbs)

            # should return the database: db6
            # Input tag is on DB, TABLE, & COLUMN level and Match_Level = TABLE_ONLY
            # Tag filter matches at TABLE level only
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.TABLE_ONLY)
            self.assertTrue(len(result.databases) == 1)
            returned_dbs.clear()
            for database in result.databases:
                returned_dbs.append(database.name)
            self.assertTrue([db6] in returned_dbs)

            # should return the database: db7
            # Input tag is on DB, TABLE, & COLUMN level and Match_Level = COLUMN_ONLY
            # Tag filter matches at TABLE level only
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.COLUMN_ONLY)
            self.assertTrue(len(result.databases) == 1)
            returned_dbs.clear()
            for database in result.databases:
                returned_dbs.append(database.name)
            self.assertTrue([db7] in returned_dbs)

            # should return the database: db5, db6, db7
            # Input tag is on DB, TABLE, & COLUMN level and Match_Level = DATABASE_PLUS
            # Tag filter matches at DB, TABLE and COLUMN level
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.DATABASE_PLUS)
            self.assertTrue(len(result.databases) == 3)
            returned_dbs.clear()
            for database in result.databases:
                returned_dbs.append(database.name)
            self.assertTrue([db5] in returned_dbs)
            self.assertTrue([db6] in returned_dbs)
            self.assertTrue([db7] in returned_dbs)

            # should return the database: db6, db7
            # Input tag is on DB, TABLE, & COLUMN level and Match_Level = TABLE_PLUS
            # Tag filter matches at TABLE and COLUMN level
            result = conn.list_databases_v2(
                tags=attributes, tag_match_level=TAttributeMatchLevel.TABLE_PLUS)
            self.assertTrue(len(result.databases) == 2)
            returned_dbs.clear()
            for database in result.databases:
                returned_dbs.append(database.name)
            self.assertTrue([db6] in returned_dbs)
            self.assertTrue([db7] in returned_dbs)

            # should return the database: db5, db6, db7
            # Input tag is on DB, TABLE, & COLUMN level and Match_Level = None or NULL
            # if no Match_Level is specified, Tag filter matches at ALL levels
            result = conn.list_databases_v2(tags=attributes)
            self.assertTrue(len(result.databases) == 3)
            returned_dbs.clear()
            for database in result.databases:
                returned_dbs.append(database.name)
            self.assertTrue([db5] in returned_dbs)
            self.assertTrue([db6] in returned_dbs)
            self.assertTrue([db7] in returned_dbs)

            # drop databases
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db5)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db6)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db7)

    def test_list_databases_access_levels(self):
        """ Test to verify that correct access levels are populated for databases and
            catalog privilege levels are correctly copied to the list of database access
            levels.
        """
        db1 = 'access_lvls_test_db1'
        db2 = 'access_lvls_test_db2'
        db3 = 'access_lvls_test_db3'
        db4 = 'access_lvls_test_db4'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # create test role and assign it to a group
            conn.execute_ddl('DROP ROLE IF EXISTS %s'% TEST_ROLE)
            conn.execute_ddl('CREATE ROLE %s'% TEST_ROLE)
            conn.execute_ddl('GRANT ROLE %s TO GROUP %s' % (TEST_ROLE, TEST_USER))
            # Grant privileges on catalog level
            conn.execute_ddl('GRANT CREATE ON CATALOG TO ROLE %s'% TEST_ROLE)
            conn.execute_ddl('GRANT SHOW ON CATALOG TO ROLE %s'% TEST_ROLE)
            conn.execute_ddl('GRANT CREATE_AS_OWNER ON CATALOG TO ROLE %s'% TEST_ROLE)
            # create database
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db1)
            conn.execute_ddl('CREATE DATABASE %s' % db1)
            # create table
            conn.execute_ddl('CREATE TABLE %s.t1(c1 int, c11 int)' % db1)
            # grant privileges on database level
            conn.execute_ddl(
                    'GRANT SELECT ON DATABASE %s to ROLE %s WITH GRANT OPTION'\
                        % (db1, TEST_ROLE))
            conn.execute_ddl('GRANT INSERT ON DATABASE %s to ROLE %s'% (db1, TEST_ROLE))

            # Test to verify that catalog access levels are copied to database within it
            ctx.enable_token_auth(token_str=TEST_USER)
            result = conn.list_databases_v2(exact_names_filter=[db1])

            for database in result.databases:
                print('Database Name: {0}'.format(database.name))
                print('Dataset Counts: {0}'.format(database.datasets_count))
                print('Access Levels: {0}'.format(database.access_levels))
                print('Has Grant Option: {0}'.format(database.has_grant))
            self.assertTrue(len(result.databases) == 1)
            for database in result.databases:
                self.assertTrue(database.name == [db1])
                self.assertTrue(database.datasets_count == 1)
                self.assertTrue(len(database.access_levels) == 4)
                assert TAccessPermissionLevel.CREATE in database.access_levels
                assert TAccessPermissionLevel.SHOW in database.access_levels
                assert TAccessPermissionLevel.SELECT in database.access_levels
                assert TAccessPermissionLevel.INSERT in database.access_levels
                assert TAccessPermissionLevel.CREATE_AS_OWNER not in \
                    database.access_levels
                assert database.has_grant

            # Switch back to admin and create a database with no table and no privileges
            # It should inherit catalog privileges (CREATE, SHOW but not CREATE_AS_OWNER)
            ctx.disable_auth()
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db2)
            conn.execute_ddl('CREATE DATABASE %s' % db2)

            ctx.enable_token_auth(token_str=TEST_USER)
            result = conn.list_databases_v2(exact_names_filter=[db2])

            self.assertTrue(len(result.databases) == 1)
            for database in result.databases:
                self.assertTrue(database.name == [db2])
                self.assertTrue(database.datasets_count == 0)
                self.assertTrue(len(database.access_levels) == 2)
                assert TAccessPermissionLevel.CREATE in database.access_levels
                assert TAccessPermissionLevel.SHOW in database.access_levels
                assert TAccessPermissionLevel.CREATE_AS_OWNER not in \
                    database.access_levels
                assert not database.has_grant

            # Test for a user having INSERT on catalog and abac-privilege on database
            # Switch back to admin
            ctx.disable_auth()
            # drop and re-create test role and assign it to a group
            # it would lose its old privileges.
            conn.execute_ddl('DROP ROLE IF EXISTS %s'% TEST_ROLE)
            conn.execute_ddl('CREATE ROLE %s'% TEST_ROLE)
            conn.execute_ddl('GRANT ROLE %s TO GROUP %s' % (TEST_ROLE, TEST_USER))
            # grant privileges on catalog level
            conn.execute_ddl('GRANT INSERT ON CATALOG TO ROLE %s'% TEST_ROLE)
            # create database
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db3)
            conn.execute_ddl('CREATE DATABASE %s' % db3)
            # create table
            conn.execute_ddl('CREATE TABLE %s.t3(c3 int, c33 int)' % db3)
            # create attribute
            conn.delete_attribute('ns_test_access_lvls', 'test_key')
            conn.create_attribute('ns_test_access_lvls', 'test_key')
            # assign attribute to the database
            conn.assign_attribute('ns_test_access_lvls', 'test_key', db3)
            # Grant abac privilege on database to test_role
            conn.execute_ddl('GRANT SELECT ON DATABASE %s HAVING attribute in (%s) \
                to ROLE %s'% (db3, 'ns_test_access_lvls.test_key', TEST_ROLE))

            ctx.enable_token_auth(token_str=TEST_USER)
            result = conn.list_databases_v2(exact_names_filter=[db3])

            self.assertTrue(len(result.databases) == 1)
            for database in result.databases:
                self.assertTrue(database.name == [db3])
                self.assertTrue(database.datasets_count == 1)
                self.assertTrue(len(database.access_levels) == 2)
                assert TAccessPermissionLevel.INSERT in database.access_levels
                assert TAccessPermissionLevel.SELECT in database.access_levels

            ctx.disable_auth()

            # create database
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db4)
            conn.execute_ddl('CREATE DATABASE %s' % db4)

            result = conn.list_databases_v2(exact_names_filter=[db4])
            # Check that admin has grant by default
            assert len(result.databases) == 1
            for database in result.databases:
                assert database.name == [db4]
                assert database.has_grant

            ctx.enable_token_auth(token_str=TEST_USER)
            result = conn.list_databases_v2(exact_names_filter=[db4])
            # Check that user hasn't grant by default
            assert len(result.databases) == 1
            for database in result.databases:
                assert database.name == [db4]
                assert not database.has_grant

            ctx.disable_auth()
            conn.execute_ddl(
                    'GRANT SELECT ON CATALOG TO ROLE %s WITH GRANT OPTION'% TEST_ROLE)

            ctx.enable_token_auth(token_str=TEST_USER)
            result = conn.list_databases_v2(exact_names_filter=[db4])
            # Check that has grant is inherited from catalog permission
            assert len(result.databases) == 1
            for database in result.databases:
                assert database.name == [db4]
                assert database.has_grant

            ctx.disable_auth()
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db1)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db2)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db3)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db4)

if __name__ == "__main__":
    unittest.main()



# Copyright 2020 Okera Inc. All Rights Reserved.
#
# Integration tests for avro in PyOkera
#
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=no-else-return
# pylint: disable=duplicate-code
# pylint: disable=line-too-long
import os
import pytest

from fastavro import reader
from okera.tests import file_format_test_common as fileFormat

LOCAL_DIR_PREFIX = '/tmp'
LOCAL_AVRO_DIR = 'test-avro'
TEST_DIR_PREFIX = 'avro-random-testing'
FILE_FORMAT = 'AVRO'

OKERA_HOME = os.environ["OKERA_HOME"]
TEST_AVRO_SAMPLE_SCHEMA_PATH = OKERA_HOME + \
                               '/tools/schema-tools/src/main/resources/test-schemas/'
TEST_DB = 'avro_test_db'
TEST_TABLE = TEST_DB + ".avro_tbl"

class AvroIntegrationTest(fileFormat.FileFormatTestBase):

    def _test_avro_files(self, local_avro_path, iters, path):

        if not os.listdir(local_avro_path):
            pytest.fail("Cannot find any generated files in test folder " + \
                         local_avro_path)

        processed_count = 0
        for file in os.listdir(local_avro_path):
            # Skip non-avro files.
            if not file.endswith(".avro"):
                continue
            else:
                processed_count += 1
                local_file_path = '%s/%s' % (local_avro_path, file)

                # Convert to json from avro file for comparison.
                output = []
                with open(local_file_path, 'rb') as f:
                    avro_reader = reader(f)
                    for record in avro_reader:
                        output.append(record)

                s3_file_path = '%s/%s/%s' % (TEST_DIR_PREFIX, path, file)
                # Upload generated avro file to s3 location.
                full_path = self._upload_s3(local_file_path, s3_file_path)

                print("Processing " + local_file_path)

                # Run comparison test using the avro file uploaded to the s3 location.
                self._test_path(full_path, output, TEST_DB, TEST_TABLE,
                                allow_missing=True, file_format=FILE_FORMAT)

        print("Total ignored samples: " + str(iters - processed_count) + "/" +
              str(iters))

    def _run_avro_random_tests(self, iters, path):
        iters = self._get_test_iters(iters)
        local_avro_path = '%s/%s/%s' % (LOCAL_DIR_PREFIX, path, LOCAL_AVRO_DIR)

        # Generate the avro file in the same temp path using java library.
        # Jar Location $OKERA_HOME/tools/schema-tools/
        self._generate_avro(local_avro_path, TEST_AVRO_SAMPLE_SCHEMA_PATH, iters)

        # Test all avro files in the folder.
        self._test_avro_files(local_avro_path, iters, path)

    def test_basic_random_schemas(self):
        test_name = 'avro-basic'
        self._cleanup(TEST_DIR_PREFIX, test_name, local_dir=LOCAL_DIR_PREFIX,
                      test_dirs=[LOCAL_AVRO_DIR])
        self._run_avro_random_tests(None, test_name)
        # Only cleanup on success, this helps with S3 consistency issues.
        self._cleanup(TEST_DIR_PREFIX, test_name, local_dir=LOCAL_DIR_PREFIX,
                      test_dirs=[LOCAL_AVRO_DIR])
